import argparse
import ast
import astunparse
import builtins
import sys

from bombast.transform import *

class Preprocess(ast.NodeVisitor):
    hardcoded = {'f_locals', 'add', 'pop'}
    ignores = set(dir(builtins)) | hardcoded

    def __init__(self):
        ast.NodeTransformer.__init__(self)
        self.mapping = {}
        self.imports = set()

    def rename(self, input):
        if input in self.ignores:
            return
        elif input in self.imports:
            return
        new_name = random.randident(4, 10)
        while new_name in self.mapping.values():
            new_name = random.randident(4, 10)
        self.mapping[input] = new_name

    def visit_Name(self, node):
        self.rename(node.id)

    def visit_FunctionDef(self, node):
        if not node.name.startswith('__'):
            self.rename(node.name)
        self.visit(node.args)
        for line in node.body:
            self.visit(line)

    def visit_ClassDef(self, node):
        self.rename(node.name)
        for base in node.bases:
            self.visit(base)
        for line in node.body:
            self.visit(line)

    def visit_Import(self, node):
        # only supports imports of the form 'import <module>'
        self.imports.add(node.names[0].name)

class Bombast(ast.NodeTransformer):
    def __init__(self, preprocess):
        super().__init__()
        self.mapping = preprocess.mapping
        self.imports = preprocess.imports

    def rename(self, input):
        return self.mapping.get(input, input)

    def visit_Expr(self, node):
        if isinstance(node.value, Str): # docstring
            return Expr(Str(s=random.randident(20, 30)))
        return Expr(self.visit(node.value))

    def visit_Num(self, node):
        return NumBombast(node).transform()

    def visit_Str(self, node):
        return StrBombast(node).transform()

    def visit_Name(self, node):
        return Name(id=self.rename(node.id), ctx=node.ctx)

    def visit_Attribute(self, node):
        if isinstance(node.value, Name) and node.value.id in self.imports:
            attr = node.attr
        else:
            attr = self.rename(node.attr)
        return Attribute(value=self.visit(node.value), attr=attr, ctx=node.ctx)

    def visit_ExceptHandler(self, node):
        return ExceptHandler(type=node.type if node.type is None else self.visit(node.type),
                             name=self.rename(node.name),
                             body=[self.visit(b) for b in node.body])

    def visit_arg(self, node):
        return arg(arg=self.rename(node.arg), annotation=node.annotation)

    def visit_keyword(self, node):
        return keyword(arg=self.rename(node.arg), value=self.visit(node.value))

    def visit_arguments(self, node):
        args = [self.visit(arg) for arg in node.args]
        kwonlyargs = [self.visit(arg) for arg in node.kwonlyargs]
        defaults, kw_defaults = node.defaults, node.kw_defaults
        if VERSION.minor < 4:
            vararg = self.rename(node.vararg)
            kwarg = self.rename(node.kwarg)
            return arguments(args, vararg, node.varargannotation,
                             kwonlyargs, kwarg, node.kwargannotation,
                             defaults, kw_defaults)
        else:
            vararg = kwarg = None
            if node.vararg is not None:
                vararg = arg(arg=self.rename(node.vararg.arg), annotation=node.vararg.annotation)
            if node.kwarg is not None:
                kwarg = arg(arg=self.rename(node.kwarg.arg), annotation=node.kwarg.annotation)
            return arguments(args, vararg, kwonlyargs, kw_defaults, kwarg, defaults)

    def visit_FunctionDef(self, node):
        name = self.rename(node.name)
        args = self.visit(node.args)
        body = [self.visit(b) for b in node.body]
        decorator_list = [self.visit(d) for d in node.decorator_list]
        return FunctionDef(name, args, body, decorator_list, node.returns)

    def visit_Global(self, node):
        return Global([self.rename(n) for n in node.names])

    def visit_Nonlocal(self, node):
        return Nonlocal([self.rename(n) for n in node.names])

    def visit_ClassDef(self, node):
        name = self.rename(node.name)
        bases = [self.visit(b) for b in node.bases]
        body = [self.visit(b) for b in node.body]
        decorator_list = [self.visit(d) for d in node.decorator_list]
        return ClassDef(name, bases,
                        node.keywords, node.starargs, node.kwargs,
                        body, decorator_list)

def main():
    parser = argparse.ArgumentParser(description='Obfuscate Python source code.')
    parser.add_argument('pyfile', help='input')
    parser.add_argument('--iters', help='number of iterations', type=int, default=1)
    parser.add_argument('--seed', help='random seed', type=int, default=0)
    args = parser.parse_args()

    random.seed(args.seed)
    if args.pyfile == '-':
        root = ast.parse(sys.stdin.read())
    else:
        root = ast.parse(open(args.pyfile, 'rb').read())

    # Choose renamings
    preprocess = Preprocess()
    preprocess.visit(root)

    bombast = Bombast(preprocess)
    for _ in range(args.iters):
        root = bombast.visit(root)
    ast.fix_missing_locations(root)
    print(astunparse.unparse(root))

if __name__ == '__main__':
    main()



import pprint
import random
import string as _string
import sys

_first_char = _string.ascii_uppercase + _string.ascii_lowercase
_charset = _string.ascii_uppercase + _string.ascii_lowercase + _string.digits*5 + '_'*20

def randident(a, b=None):
    length = None
    try:
        length = random.randrange(a, b)
    except:
        pass
    if length is None or b is None:
        length = a
    return random.choice(_first_char) + ''.join(random.choice(_charset) for _ in range(length-1))
random.randident = randident

def debug(*args):
    print(*args, file=sys.stderr)

VERSION = sys.version_info



from ast import Num, Str, List, Tuple, Set, Dict # NameConstant
from ast import Name, Load, Store, Del, Starred
from ast import Expr, UnaryOp, UAdd, USub, Not, BinOp
from ast import Add, Sub, Mult, Div, FloorDiv, Mod, Pow
from ast import BoolOp, And, Or, Compare, Eq, NotEq, Lt, LtE, Gt, GtE, Is, IsNot, In, NotIn
from ast import Call, keyword, IfExp, Attribute
from ast import Subscript, Index, Slice
from ast import ListComp, SetComp, DictComp, GeneratorExp, comprehension
from ast import Assign, AugAssign, Raise, Assert, Delete, Pass
from ast import Import, ImportFrom, alias
from ast import If, For, While, Break, Continue, ExceptHandler
from ast import With # withitem
from ast import FunctionDef, Lambda, arguments, arg, Return, Yield, Global, Nonlocal
from ast import ClassDef

from bombast.utils import *

class Transformation(object):
    def __init__(self, *fns):
        self.fns = fns

    def transform(self, input):
        if self.fns:
            return random.choice(self.fns)(input)
        return input


class PrimitiveBombast(object):
    def __init__(self, node):
        self.node = node

    def transform(self, input):
        return input


class RenameBombast(PrimitiveBombast):
    def __init__(self, node, bombast):
        super().__init__(node)
        self.bombast = bombast


class StrBombast(PrimitiveBombast):
    def transform(self):
        s = self.node.s
        if len(s) == 0:
            return self.zero.transform(self.node)
        elif len(s) == 1:
            return self.one.transform(self.node)
        else:
            return self.many.transform(self.node)

    def zero_Constructor(node): # '' -> str()
        return Call(func=Name(id='str', ctx=Load()),
                    args=[], keywords=[], starargs=None, kwargs=None)
    def zero_Identity(node): # '' -> ''
        return node
    zero = Transformation(zero_Constructor, zero_Identity)

    def one_Ordinal(node): # 'a' -> chr(97)
        return Call(func=Name(id='chr', ctx=Load()),
                    args=[Num(ord(node.s))], keywords=[], starargs=None, kwargs=None)
    def one_Identity(node): # 'a' -> 'a'
        return node
    one = Transformation(one_Ordinal, one_Identity)

    def many_Split(node): # 'hello' -> 'h' + 'ello' (with randomly chosen cut)
        s = node.s
        i = random.randrange(len(s))
        return BinOp(
            left=Str(s=s[:i]),
            right=Str(s=s[i:]),
            op=Add()
        )
    many = Transformation(many_Split)


class NumBombast(PrimitiveBombast):
    def transform(self):
        n = self.node.n
        if not n:
            return self.zero.transform(self.node)
        elif isinstance(n, int):
            return self.int.transform(self.node)
        else:
            return self.float.transform(self.node)

    def zero_Multiplier(node): # 0 -> int(n * 0)
        return Call(
            func=Name(id='int', ctx=Load()),
            args=[BinOp(left=Num(n=random.random()), right=Num(n=0), op=Mult())],
            keywords=[], starargs=None, kwargs=None
        )
    def zero_Identity(node):
        return node
    zero = Transformation(zero_Multiplier, zero_Identity)

    def int_Split(node, range=100): # n -> (n-s) + (s)
        s = random.randint(-range, range)
        return BinOp(
            left=Num(n=node.n-s),
            right=Num(n=s),
            op=Add()
        )
    int = Transformation(int_Split)

    def float_Split(node): # n -> (n-s) + (s)
        s = random.random()
        return BinOp(
            left=Num(n=node.n-s),
            right=Num(n=s),
            op=Add()
        )
    float = Transformation(float_Split)


class ImportBombast(RenameBombast):
    # import sys -> sys = __import__('sys', globals(), locals(), [], 0)
    one = Transformation(
        lambda n: Assign(
            targets=[Name(id=n.names[0].name, ctx=Store())],
            value=Call(func=Name(id='__import__', ctx=Load()),
                       args=[
                           Str(s=n.names[0].name),
                           Call(func=Name(id='globals', ctx=Load()),
                                args=[], keywords=[], starargs=None, kwargs=None),
                           Call(func=Name(id='locals', ctx=Load()),
                                args=[], keywords=[], starargs=None, kwargs=None),
                           List(elts=[], ctx=Load()),
                           Num(n=0)
                       ],
                       keywords=[], starargs=None, kwargs=None))
        )

    def transform(self):
        num_imports = len(self.node.names)
        if num_imports == 1:
            return self.one.transform(self.node)
        else:
            return self.node



from setuptools import setup

setup(
    name='bombast',
    version='v0.1.7',
    description='An obfuscator for Python source code that manipulates the AST.',
    url='https://github.com/brianhou/bombast',
    author='Brian Hou',
    packages=['bombast'],
    install_requires=['astunparse'],
    entry_points={
        'console_scripts': [
            'bombast=bombast.__init__:main',
        ],
    }
)
